#!/usr/bin/perl
#
# Partial_gene_sorter_v1.02.pl
#
# Copyright Jonathan Klassen, 2011
# This program is distributed under the GNU General Public License v3. See the bottom of the accompanying README file for terms of this license.
#
# Please cite: Klassen and Currie (2011) BMC Genomics
#
# Uses pair-wise BLASTs to assemble broken genes based on their BLAST homology to other genes in the target BLAST database
# Criterion for best hit: AAI similarity over some percentage of the partial sequence BLAST hit
# Defaults are set as the best values determined in our paper (Klassen and Currie BMC Genomics 2011)

use warnings;

######################################################################################################################
# processes input arguments
######################################################################################################################

$usage = "Partial_gene_sorter_v1.02.pl

Copyright Jonathan Klassen, 2011
This program is distributed under the GNU General Public License v3. See the bottom of the accompanying README file for terms of this license.

Please cite: Klassen and Currie (2011) BMC Genomics

USAGE:
	-i 	input list of paths to the protein fasta files to be analyzed					REQUIRED
	-o 	name for the output file listing the orf names for each fragment set				DEFAULT: list.out
	-p 	name for the output file listing the detailed descriptions for each fragment set		DEFAULT: details.out
	-a 	the minimum fractional identity between a hit and its reference sequence			DEFAULT: 0.2
	-b	the minimum fractional homologous length between a hit and its reference sequence		DEFAULT: 0.6
	-c	the maximum difference in fractional identities between hits in a set and their reference	DEFAULT: 0.5
	-d	the maximum number of overlapping amino acids between fragments					DEFAULT: 0
	-e	the number of cores to use for BLASTing								DEFAULT: 1\n";

if (scalar @ARGV % 2 > 0){ die $usage } # the input argument format is <flag> <statement without spaces>

for ($a = 0; $a < scalar @ARGV; $a = $a + 2){
	if ($ARGV[$a] eq "-i"){
		$input_file = $ARGV[$a + 1];
	}
	elsif ($ARGV[$a] eq "-o"){
		$list_output_file = $ARGV[$a + 1];
	}
	elsif ($ARGV[$a] eq "-p"){
		$descriptions_output_file = $ARGV[$a + 1];
	}
	elsif ($ARGV[$a] eq "-a"){
		$ID = $ARGV[$a + 1];
		die $usage unless ($ID =~ /^\d+$/); # fractional identity values must be numeric
	}
	elsif ($ARGV[$a] eq "-b"){
		$ID_length = $ARGV[$a + 1];
		die $usage unless ($ID_length =~ /^\d+$/); # fractional homologous length values must be numeric
	}
	elsif ($ARGV[$a] eq "-c"){
		$AAI_difference = $ARGV[$a + 1];
		die $usage unless ($AAI_difference =~ /^\d+$/); # identity difference values must be numeric
	}
	elsif ($ARGV[$a] eq "-d"){
		$aa_overlap = $ARGV[$a + 1];
		die $usage unless ($aa_overlap =~ /^\d+$/); # sequence overlap difference values must be numeric
	}
	elsif ($ARGV[$a] eq "-e"){
		$cores = $ARGV[$a + 1];
		die $usage unless ($cores =~ /^\d+$/); # number of cores to use must be numeric
	}
	else { die $usage } # mystery flags are not allowed
}

# sets defaults as necessary

die $usage unless ($input_file); # the input file name is manditory
unless ($list_output_file){ $list_output_file = "list.out" }
unless ($descriptions_output_file){ $descriptions_output_file = "details.out" }
unless ($ID){ $ID = 0.2 }
unless ($ID_length){ $ID_length = 0.6 }
unless ($AAI_difference){ $AAI_difference = 0.5 }
unless ($aa_overlap){ $aa_overlap = 0 }
unless ($cores){ $cores = 1 }

######################################################################################################################
# loads input file list
######################################################################################################################

# stores all input files in the "filenames" array

open (INLIST, $input_file) or die "Cannot open $input_file";
while (<INLIST>){
	s/\s*$//;
	push @filenames, $_;
}
close INLIST;

######################################################################################################################
# iterates through all pair-wise combinations of the input files, BLASTing and assembling fragment sets as appropriate
######################################################################################################################

$hit_counter = -1; #counter for eventual output array of all protein groups
for ($aa = 0; $aa < scalar @filenames; $aa++){

	# extracts the genome name from the input file path, removing the directory if necessary. Assumes a *.* file format.

	if ($filenames[$aa] =~ /\/?((\w|-)+)\..*$/){ $name1 = $1 }
	else { die "Cannot extract the strain name from $filenames[$aa]. Does $filenames[$aa] have a *.* format? Does the name contain characters other than letters, numbers, underscores or dashes?" }

	for ($bb = $aa + 1; $bb < scalar @filenames; $bb++){

		if ($filenames[$bb] =~ /\/?((\w|-)+)\..*$/){ $name2 = $1 }
		else { die "Cannot extract strain name from $filenames[$bb]. Does $filenames[$bb] have a *.* format? Does the name contain characters other than letters, numbers, underscores or dashes?" }

		#############################################################################################################		
		# BLASTp using the parameters from Konstantinidis and Tiedje (2005)
		# Only does BLAST if the blast output is not found in the directory from which the script is run
		# Only makes BLAST databases if they are not found in the directory where the BLAST input sequence is located
		############################################################################################################		

		unless (-e "$name1\_vs_$name2.blastp" and -e "$name2\_vs_$name1.blastp"){
			print "BLASTing $name1 vs $name2...";
			unless (-e "$filenames[$aa].pin"){ system ("formatdb -i $filenames[$aa]") }
			unless (-e "$filenames[$bb].pin"){ system ("formatdb -i $filenames[$bb]") }				
			system ("blastall -i $filenames[$aa] -d $filenames[$bb] -a $cores -p blastp -b 20 -v 20 -e 1e-05 -F F > $name1\_vs_$name2.blastp");
			print "done\nBLASTing $name2 vs $name1...";
			system ("blastall -i $filenames[$bb] -d $filenames[$aa] -a $cores -p blastp -b 20 -v 20 -e 1e-05 -F F > $name2\_vs_$name1.blastp"); 
			print "done\n";
		}

		#############################################################################################################		
		# Extracts information from the BLAST of genome #1 vs genome #2
		# Assumes all OPartial_gene_sorter_v1.02.pl
#
# Copyright Jonathan Klassen, 2011
# This program is distributed under the GNU General Public License v3. See the bottom of the accompanying README file for terms of this license.
#
# Please cite: Klassen and Currie (2011) BMC GenomicsRFs will have in their header "partial=00", "partial=01", "partial=10" or "partial=11" to
		# represent complete, 5'-truncated, 3'-truncated or 5' and 3'-truncated orfs, as Prodigal does during 
		# gene calling.
		############################################################################################################		

		print "Extracting information from BLAST1 ($name1)...";

		%inblast1 = %inblast2 = @querystart = @subjectstart = @queryend = @subjectend = ();		
		open (INBLAST1, "$name1\_vs_$name2.blastp") or die "Cannot open $name1\_vs_$name2.blastp ";
		while (<INBLAST1>){
			if (/^Query= ((\w|-|\.)+)\b/){ 
				$ref_name = $1;
				$found_query = 1;
				$hit_number = -1;
			}
			if (/(partial=\d+)/ and $found_query == 1){
				$inblast1{$ref_name}{'ref_fragment'} = $1;
				$found_query = 0;
			}
			elsif (/\((\d+,*\d*) letters\)/){
				$inblast1{$ref_name}{'ref_length'} = $1;
				$inblast1{$ref_name}{'ref_length'} =~ s/,//g;
			}
			elsif (/>((\w|-|\.)+)\b/){
				$hit_number++;
				$hitcounter = 0;
				@querystart = @queryend = @subjectstart = @subjectend = ();
				push @{$inblast1{$ref_name}{'hitname'}}, $1;
			}
			elsif (/Length = (\d+,*\d*)/){
				push @{$inblast1{$ref_name}{'hit_length'}}, $1;
				$inblast1{$ref_name}{'hit_length'}[$#{$inblast1{$ref_name}{'hit_identities'}}] =~ s/,//g;
			}
			elsif (/Identities = (\d+)\/(\d+) \(\d+\%\)/){
				$hitcounter++; #increments because only the first homologous region is accepted (blast parameter -F F)
				$y = $1;
				$z = $2;
				if ($hitcounter == 1){
					push @{$inblast1{$ref_name}{'hit_identities'}}, $y;
					push @{$inblast1{$ref_name}{'hit_alignment'}}, $z;	
					push @{$inblast1{$ref_name}{'AAI'}}, ($inblast1{$ref_name}{'hit_identities'}[$#{$inblast1{$ref_name}{'hit_identities'}}] / 
						$inblast1{$ref_name}{'hit_alignment'}[$#{$inblast1{$ref_name}{'hit_alignment'}}]);
				}
			}
			elsif (/^Query: (\d+)/){
				if ($hitcounter == 1){ #only harvests data from the first homologous region
					push @querystart, $1;
					/(\d+)$/;
					push @queryend, $1;
					$inblast1{$ref_name}{'query_start'}[$hit_number] = $querystart[0];
					$inblast1{$ref_name}{'query_end'}[$hit_number] = $queryend[$#queryend];
				}
			}
			elsif (/^Sbjct: (\d+)/){
				if ($hitcounter == 1){ #only harvests data from the first homologous fregion
					push @subjectstart, $1;
					/(\d+)$/;
					push @subjectend, $1;
					$inblast1{$ref_name}{'subject_start'}[$hit_number] = $subjectstart[0];
					$inblast1{$ref_name}{'subject_end'}[$hit_number] = $subjectend[$#subjectend];
				}
			}
		}
		close INBLAST1;
		
		#############################################################################################################		
		# Extracts information from the BLAST of genome #1 vs genome #2
		# Assumes all ORFs will have in their header "partial=00", "partial=01", "partial=10" or "partial=11" to
		# represent complete, 5'-truncated, 3'-truncated or 5' and 3'-truncated orfs, as Prodigal does during 
		# gene calling.
		############################################################################################################		

		print "done\nExtracting information from BLAST2 ($name2)...";

		@querystart = @subjectstart = @queryend = @subjectend = ();

		open (INBLAST2, "$name2\_vs_$name1.blastp") or die "Cannot open $name2\_vs_$name1.blastp ";
		while (<INBLAST2>){
			if (/^Query= ((\w|-|\.)+)\b/){ 
				$ref_name = $1;
				$found_query = 1;
				$hit_number = -1;
			}
			if (/(partial=\d+)/ and $found_query == 1){
				$inblast2{$ref_name}{'ref_fragment'} = $1;
				$found_query = 0;
			}
			elsif (/\((\d+,*\d*) letters\)/){
				$inblast2{$ref_name}{'ref_length'} = $1;
				$inblast2{$ref_name}{'ref_length'} =~ s/,//g;
			}
			elsif (/>((\w|-|\.)+)\b/){
				$hit_number++;
				$hitcounter = 0;
				@querystart = @queryend = @subjectstart = @subjectend = ();
				push @{$inblast2{$ref_name}{'hitname'}}, $1;
			}
			elsif (/Length = (\d+,*\d*)/){
				push @{$inblast2{$ref_name}{'hit_length'}}, $1;
				$inblast2{$ref_name}{'hit_length'}[$#{$inblast2{$ref_name}{'hit_identities'}}] =~ s/,//g;
			}
			elsif (/Identities = (\d+)\/(\d+) \(\d+\%\)/){
				$hitcounter++; #increments because only the first homologous region is accepted (blast parameter -F F)
				$y = $1;
				$z = $2;
				if ($hitcounter == 1){
					push @{$inblast2{$ref_name}{'hit_identities'}}, $y;
					push @{$inblast2{$ref_name}{'hit_alignment'}}, $z;######################################################################################################################
# iterates through all pair-wise combinations of the input files, BLASTing and assembling fragment sets as appropriate
######################################################################################################################
	
					push @{$inblast2{$ref_name}{'AAI'}}, ($inblast2{$ref_name}{'hit_identities'}[$#{$inblast2{$ref_name}{'hit_identities'}}] / 
						$inblast2{$ref_name}{'hit_alignment'}[$#{$inblast2{$ref_name}{'hit_alignment'}}]);
				}
			}
			elsif (/^Query: (\d+)/){
				if ($hitcounter == 1){ #only harvests data from the first homologous region
					push @querystart, $1;
					/(\d+)$/;
					push @queryend, $1;
					$inblast2{$ref_name}{'query_start'}[$hit_number] = $querystart[0];
					$inblast2{$ref_name}{'query_end'}[$hit_number] = $queryend[$#queryend];
				}
			}
			elsif (/^Sbjct: (\d+)/){
				if ($hitcounter == 1){ #only harvests data from the first homologous fregion
					push @subjectstart, $1;
					/(\d+)$/;
					push @subjectend, $1;
					$inblast2{$ref_name}{'subject_start'}[$hit_number] = $subjectstart[0];
					$inblast2{$ref_name}{'subject_end'}[$hit_number] = $subjectend[$#subjectend];
				}
			}
		}
		close INBLAST2;
		print "done\n";

		############################################################################################################		
		# Determines matching fragment sets
		############################################################################################################

		%match = ();
		foreach $ref_name (sort keys %inblast1){
			$end_counter = $middle_counter = $temp_inblast1_best_hit_array_key = 0;

			#next sequence if no hits to the genome #1 query are found in genome #2

			next if (!$inblast1{$ref_name}{'hitname'}[0]);

			#next if the query sequence in genome #1 is not partial 

			next if ($inblast1{$ref_name}{'ref_fragment'} eq "partial=00");

			#finds the best hit in genome #2 to the genome #1 query sequence
			#scans through all hits, and finds the best hit > $ID and with alignment > $ID_length

			$temp_best_ref_AAI = $temp_inblast1_best_hit_array_key = 0;
			foreach $a (0..$#{$inblast1{$ref_name}{'hitname'}}){
				$b = $inblast1{$ref_name}{'hitname'}[$a];
				next if (!$inblast2{$b}{'ref_fragment'});
				next if ($inblast2{$b}{'ref_fragment'} ne "partial=00");
				if ($inblast1{$ref_name}{'AAI'}[$a] > $temp_best_ref_AAI and $inblast1{$ref_name}{'AAI'}[$a] > $ID and 
					($inblast1{$ref_name}{'hit_alignment'}[$a] / $inblast1{$ref_name}{'ref_length'}) > $ID_length){
					$temp_best_ref_AAI = $inblast1{$ref_name}{'AAI'}[$a];
					$temp_inblast1_best_hit_array_key = $a;
				}
			}
			next if (!$temp_best_ref_AAI);

			#cycles through reciprocal hits in genome #2 back to genome #1; selects matching fragments with the highest AAIs > $ID and having alignments
			# > $ID_length; only keeps 2 ends, one of which is the query
			
			$b = $inblast1{$ref_name}{'hitname'}[$temp_inblast1_best_hit_array_key];

			foreach $c (0..$#{$inblast2{$b}{'hitname'}}){
				$next_counter = 0;

				#next if the reciprocal hit is not a fragment

				$d = $inblast2{$b}{'hitname'}[$c];
				next if ($inblast1{$d}{'ref_fragment'} eq "partial=00");

				#next if the AAI of the hit is <= the cutoff threshold $ID

				next if ($inblast2{$b}{'AAI'}[$c] <= $ID);

				#next if the aligned length of the hit is <= the cutoff threshold $ID_length
	
				next if (($inblast2{$b}{'hit_alignment'}[$c] / $inblast2{$b}{'hit_length'}[$c]) <= $ID_length);

				#next if the difference in AAI between the original query match and the current AAI > the cutoff threshold $AAI_difference

				next if ((abs($inblast1{$ref_name}{'AAI'}[$temp_inblast1_best_hit_array_key] - $inblast2{$b}{'AAI'}[$c])) > $AAI_difference);

				#adds reciprocal hits in genome #1 back to genome #1 that pass all criteria to %match

				#if the hit to be added is an end fragment, checks to ensure that there has not been an end fragment not original query from genome #1
				#	with a higher AAI already i.e., only 2 ends allowed, and the ones with the highest AAIs are selected preferentially.

				if ($inblast1{$d}{'ref_fragment'} eq "partial=01" or $inblast1{$d}{'ref_fragment'} eq "partial=10"){
					if ($end_counter == 2){
						%temp_best_difference_key = ();
						foreach $e (keys %{$match{$ref_name}}){ #searches through fragement type keys 
							next if ($match{$ref_name}{$e} =~ /middle/); #skips middle fragments

							#Marks to skip if the hit AAI matches the query AAI > the match AAI does
						
							if ((abs($inblast1{$ref_name}{'AAI'}[$temp_inblast1_best_hit_array_key] - $inblast2{$b}{'AAI'}[$c])) > 
								(abs($inblast1{$ref_name}{'AAI'}[$temp_inblast1_best_hit_array_key] - $match{$ref_name}{$e}{'AAI'}))){
								$next_counter = 1;
							} 

							#Otherwise flags match as better, removes previous indication to skip; this step need because it is possible
							#	to have 2 ends + the query a "middle", and the new end might only be better than one of them
			
							else {
								$temp_best_difference_key{$e} = abs($inblast1{$ref_name}{'AAI'}[$temp_inblast1_best_hit_array_key] - $match{$ref_name}{$e}{'AAI'});
								$next_counter = 0;
							}
						}

						#determines which of the existing matching ends to skip

						unless ($next_counter == 1){
							if (!$temp_best_difference_key{'end1'}){$end_name = "end2"}
							elsif (!$temp_best_difference_key{'end2'}){$end_name = "end1"}
							elsif ($temp_best_difference_key{'end1'} >= $temp_best_difference_key{'end2'}){$end_name = "end1"}
							elsif ($temp_best_difference_key{'end1'} < $temp_best_difference_key{'end2'}){$end_name = "end2"}
						}
					}
		
					#If 2 ends have not been previously found, set the name to add the current end fragment

					else {
						$end_counter++;
						$end_name = "end".$end_counter;
					}
				}

				#Sets the name to add the current middle sequence fragment

				else {
					$middle_counter++;
					$end_name = "middle".$middle_counter;
				}

				#Skips if the end sequence is not being replaced; see above

				next if ($next_counter == 1);

				#Adds fragment data

				$match{$ref_name}{$end_name}{'hitname'} = $inblast2{$b}{'hitname'}[$c];
				$match{$ref_name}{$end_name}{'fragment'} = $inblast1{$d}{'ref_fragment'};
				$match{$ref_name}{$end_name}{'hit_length'} = $inblast2{$b}{'hit_length'}[$c];
				$match{$ref_name}{$end_name}{'hit_alignment'} = $inblast2{$b}{'hit_alignment'}[$c];
				$match{$ref_name}{$end_name}{'hit_identities'} = $inblast2{$b}{'hit_identities'}[$c];
				$match{$ref_name}{$end_name}{'query_start'} = $inblast2{$b}{'query_start'}[$c];
				$match{$ref_name}{$end_name}{'query_end'} = $inblast2{$b}{'query_end'}[$c];
				$match{$ref_name}{$end_name}{'subject_start'} = $inblast2{$b}{'subject_start'}[$c];
				$match{$ref_name}{$end_name}{'subject_end'} = $inblast2{$b}{'subject_end'}[$c];
				$match{$ref_name}{$end_name}{'AAI'} = $inblast2{$b}{'AAI'}[$c];
				$match{$ref_name}{$end_name}{'genome_2_hit'} = $inblast1{$ref_name}{'hitname'}[$temp_inblast1_best_hit_array_key];
			}							

		#Once best hits are recovered, a series of other tests are conducted to ensure fidelity

			#Determines if the query sequence is recovered; otherwise deletes the hash element
			
			$found_ref_name = 0;
			foreach $f (sort keys %{$match{$ref_name}}){
				if ($ref_name eq $match{$ref_name}{$f}{'hitname'}){$found_ref_name = 1}
			}
			if ($found_ref_name == 0){
				delete $match{$ref_name};
				next;
			}
			
			#Determines if the starts and stops for each "assembled" protein overlap by less than 25aa. If they do, reject the assembly.
			#Makes arrays of all starts and stops. With the sorted keys, the query sequence is at the start and its corresponding fragment is at the
			#other end.
 
			@query_starts = @query_ends = ();
			foreach $f (sort keys %{$match{$ref_name}}){ 
				if ($match{$ref_name}{$f}{'query_start'} < $match{$ref_name}{$f}{'query_end'}){ #ifs accommodate for backwards alignment
					push @query_starts, $match{$ref_name}{$f}{'query_start'};
					push @query_ends, $match{$ref_name}{$f}{'query_end'};
				}
				elsif ($match{$ref_name}{$f}{'query_start'} > $match{$ref_name}{$f}{'query_end'}){
					push @query_starts, $match{$ref_name}{$f}{'query_end'};
					push @query_ends, $match{$ref_name}{$f}{'query_start'};
				}
			}
			@temp = sort {$a <=> $b} @query_starts;
			@query_starts = @temp;
			@temp = sort {$a <=> $b} @query_ends;
			@query_ends = @temp;

			#Deletes hash element if only match (i.e. to itself, due to the criteria above) found

			if (!$query_starts[1]){
				delete $match{$ref_name};
				next;
			}

			#Looks for overlaps having less than a maximum of 25 amino acids; if found deletes hash element		

			for ($g = 1; $g < scalar @query_starts; $g++){ 
				if ($query_starts[$g] <= ($query_ends[$g-1] - $aa_overlap)){
					delete $match{$ref_name};
					last;				
				}
			}
		}

		%global_found = ();
		foreach $ref_name (sort keys %match){ #foreach hit
			$no_match = $last_counter = $temp_avg_match_AAI = $temp_avg_ref_AAI = $not_found = $last_counter = $better_counter = 0;
			%found_hit_counters = @temp_ref_AAIs = ();
			$d = ''; 

			#Because sets of matches are compared, good matches can be stored in %global_found out of order. Next if this has happened.

			next if ($global_found{$ref_name});

			#Finds each partner for each match, and determines if that partner also reciprocally found the first match.
			#Counters are used here, because if there are any incongruencies between the matching sets, the series of sets is flagged.

			foreach $f (sort keys %{$match{$ref_name}}){ #foreach part 
				$found = 0;
				$a = $match{$ref_name}{$f}{'hitname'};
				foreach $b (sort keys %{$match{$a}}){ #foreach part of the reciprical
					if ($match{$ref_name}{$f}{'hitname'} eq $match{$a}{$b}{'hitname'}){#check for match
						$found = 1;

						#Prevents matching same hit twice per pairwise genome comparison

						$global_found{$match{$ref_name}{$f}{'hitname'}} = " "; 
					}
				}
				if ($found == 0){#if no match
					$no_match = 1;
				}
			}
			next if ($no_match == 1);

			if ($no_match == 0){#if match

				#evaluates if the current match comprises hits from multiple previously identified sets

				foreach $f (sort keys %{$match{$ref_name}}){#foreach part
					$c = $match{$ref_name}{$f}{'hitname'};
					if ($global_index{$c}){
						$found_hit_counters{$global_index{$c}} = " ";
					}
				}

				#if the current match comprises hits from multiple previously identified sets

				if (scalar keys %found_hit_counters > 1){

					#calculates the average AAI for the match hit set
			
					$temp_avg_match_AAI = 0;
					foreach $f (sort keys %{$match{$ref_name}}){
						$temp_avg_match_AAI += $match{$ref_name}{$f}{'AAI'};
					}
					$temp_avg_match_AAI = $temp_avg_match_AAI / (scalar keys %{$match{$ref_name}});

					#calculates the average AAI of teach previously identified hit set

					foreach $hit_counter (keys %found_hit_counters){
						for $evidence_counter (0..$#{$global_match[$hit_counter]}){
							$temp_avg_ref_AAI = 0;
							for $part_counter (0..$#{$global_match[$hit_counter][$evidence_counter]}){
								$temp_avg_ref_AAI += $global_match[$hit_counter][$evidence_counter][$part_counter]{'AAI'};
							}
							push @temp_ref_AAIs, ($temp_avg_ref_AAI / (scalar @{$global_match[$hit_counter][$evidence]}));
						 	@temp_ref_AAIs = sort {$a <=> $b} @temp_ref_AAIs;
							if ($temp_ref_AAIs[$#temp_ref_AAIs] >= $temp_avg_match_AAI){
								$last_counter = 1;
								last;
							}
						}
						last if ($last_counter == 1);
					}
					last if ($last_counter == 1);

					#if all previously identified hit sets have average AAIs worse than that of the current match set, delete the old sets
					#	referenced in @global_match and %global_index

					foreach $hit_counter (keys %found_hit_counters){
						for $evidence_counter (0..$#{$global_match[$hit_counter]}){
							for $part_counter (0..$#{$global_match[$hit_counter][$evidence_counter]}){
								$global_index{$global_match[$hit_counter][$evidence_counter][$part_counter]{'hitname'}} = ();
							}
						}
						$global_match[$hit_counter] = ();
					}
					next if ($last_counter == 1);
				}

				#if the current match shares parts with a previously found set, ensures that all of the previously found set is included in the
				#	current set

				elsif (scalar keys %found_hit_counters == 1){
					foreach (keys %found_hit_counters){$d = $_}
					for $evidence_counter (0..$#{$global_match[$d]}){
						for $part_counter (0..$#{$global_match[$d][$evidence_counter]}){
							$temp_match = 0;
							foreach $f (keys %{$match{$ref_name}}){
								if ($match{$ref_name}{$f}{'hitname'} eq 
									$global_match[$d][$evidence_counter][$part_counter]{'hitname'}){$temp_match = 1};
							}
							if ($temp_match == 0){
								$not_found = 1;
								$last_counter = 1;
								last;
							}
						}
						last if ($last_counter == 1);
					}

					#if there as a part of the previously found set that is not found in the current match set, the average AAI for both sets are
					#	compared and the best set is carried forward

					if ($not_found == 1){
						$temp_avg_match_AAI = 0;
						foreach $f (keys %{$match{$ref_name}}){
							$temp_avg_match_AAI += $match{$ref_name}{$f}{'AAI'};
						}
						$temp_avg_match_AAI = ($temp_avg_match_AAI / scalar keys %{$match{$ref_name}});
						for $evidence_counter (0..$#{$global_match[$d]}){
							$temp_avg_ref_AAI = 0;
							for $part_counter (0..$#{$global_match[$d][$evidence_counter]}){
								$temp_avg_ref_AAI += $global_match[$d][$evidence_counter][$part_counter]{'AAI'};
							}
							push @temp_ref_AAIs, ($temp_avg_ref_AAI / (scalar @{$global_match[$d][$evidence]}));
						 	@temp_ref_AAIs = sort {$a <=> $b} @temp_ref_AAIs;
							if ($temp_ref_AAIs[$#temp_ref_AAIs] >= $temp_avg_match_AAI){
								$better_counter = 1;
								last;
							}
						}
						last if ($better_counter == 1);

						#if all previously identified hit sets have average AAIs worse than that of the current match set, delete the old sets
						#	referenced in @global_match and %global_index

						for $evidence_counter (0..$#{$global_match[$d]}){
							for $part_counter (0..$#{$global_match[$d][$evidence_counter]}){
								$global_index{$global_match[$d][$evidence_counter][$part_counter]{'hitname'}} = ();
							}
						}
						$global_match[$d] = ();
						$d = '';
					}
				}

				next if ($better_counter == 1);

				#at this point, if the loop is still running the match set is to be added to @global_match
				#if old @global_match and %global_index values were to be deleted, they already would have been

				#$d is only set if the one $hit_counter is found in %global_index, and only carried through to the step adding the match if new parts
				#	are to be added $global_match[$hit_counter]; otherwise it will remain undefined

				if ($d) {
					$hit_counter = $d;
					$evidence_counter = scalar @{$global_match[$hit_counter]};
				}
				else {
					$hit_counter = scalar @global_match;
					$evidence_counter = 0;
				}

				#adds the data in the %match set to @global_match
	
				$part_counter = -1;
				foreach $f (sort keys %{$match{$ref_name}}){
					$part_counter++;
					%{$global_match[$hit_counter][$evidence_counter][$part_counter]} = (
						'hitname' => $match{$ref_name}{$f}{'hitname'},
						'fragment' => $match{$ref_name}{$f}{'fragment'},
						'AAI' => $match{$ref_name}{$f}{'AAI'},
						'hit_length' => $match{$ref_name}{$f}{'hit_length'},
						'hit_alignment' => $match{$ref_name}{$f}{'hit_alignment'},
						'hit_identities' => $match{$ref_name}{$f}{'hit_identities'},
						'query_start' => $match{$ref_name}{$f}{'query_start'},
						'query_end' => $match{$ref_name}{$f}{'query_end'},
						'subject_start' => $match{$ref_name}{$f}{'subject_start'},
						'subject_end' => $match{$ref_name}{$f}{'subject_end'},
						'genome_2_hit' => $match{$ref_name}{$f}{'genome_2_hit'},
					);

					#if needed, sets the index reference for this part

					$c = $match{$ref_name}{$f}{'hitname'};
					unless ($global_index{$c}){
						$global_index{$c} = $hit_counter;
					} 						
				}
			}
		}	

		#Same for genome #2

		%match = ();
		foreach $ref_name (sort keys %inblast2){
			$end_counter = $middle_counter = $temp_inblast2_best_hit_array_key = 0;

			#next sequence if no hits to the query are found in genome #1

			next if (!$inblast2{$ref_name}{'hitname'}[0]);

			#next if the query sequence in genome #2 is not partial 

			next if ($inblast2{$ref_name}{'ref_fragment'} eq "partial=00");

			#finds the best hit in genome #1 to the genome #2 query sequence
			#scans through all hits, and finds the best hit > $ID and with alignment > $ID_length

			$temp_best_ref_AAI = $temp_inblast2_best_hit_array_key = 0;
			foreach $a (0..$#{$inblast2{$ref_name}{'hitname'}}){
				$b = $inblast2{$ref_name}{'hitname'}[$a];
				next if ($inblast1{$b}{'ref_fragment'} ne "partial=00");
				if ($inblast2{$ref_name}{'AAI'}[$a] > $temp_best_ref_AAI and $inblast2{$ref_name}{'AAI'}[$a] > $ID and 
					($inblast2{$ref_name}{'hit_alignment'}[$a] / $inblast2{$ref_name}{'ref_length'}) > $ID_length){
					$temp_best_ref_AAI = $inblast2{$ref_name}{'AAI'}[$a];
					$temp_inblast2_best_hit_array_key = $a;
				}
			}
			next if (!$temp_best_ref_AAI);

			#cycles through reciprocal hits in genome #1 back to genome #2; selects matching fragments with the highest AAIs > $ID and having alignments
			# > $ID_length; only keeps 2 ends, one of which is the query
			
			$b = $inblast2{$ref_name}{'hitname'}[$temp_inblast2_best_hit_array_key];
			foreach $c (0..$#{$inblast1{$b}{'hitname'}}){
				$next_counter = 0;
				next if (!$inblast1{$b}{'hitname'}[0]); # next if no reciprocal hit

				#next if the reciprocal hit is not a fragment

				$d = $inblast1{$b}{'hitname'}[$c];
				next if (!$inblast2{$d}{'ref_fragment'});
				next if ($inblast2{$d}{'ref_fragment'} eq "partial=00");

				#next if the AAI of the hit is <= the cutoff threshold $ID

				next if ($inblast1{$b}{'AAI'}[$c] <= $ID);

				#next if the aligned length of the hit is <= the cutoff threshold $ID_length
	
				next if (($inblast1{$b}{'hit_alignment'}[$c] / $inblast1{$b}{'hit_length'}[$c]) <= $ID_length);

				#next if the difference in AAI between the original query match and the current AAI > the cutoff threshold $AAI_difference

				next if ((abs($inblast2{$ref_name}{'AAI'}[$temp_inblast2_best_hit_array_key] - $inblast1{$b}{'AAI'}[$c])) > $AAI_difference);

				#adds reciprocal hits in genome #1 back to genome #2 that pass all criteria to %match

				#if the hit to be added is an end fragment, checks to ensure that there has not been an end fragment not original query from genome #2
				#	with a higher AAI already i.e., only 2 ends allowed, and the ones with the highest AAIs are selected preferentially.

				if ($inblast2{$d}{'ref_fragment'} eq "partial=01" or $inblast2{$d}{'ref_fragment'} eq "partial=10"){
					if ($end_counter == 2){
						%temp_best_difference_key = ();
						foreach $e (keys %{$match{$ref_name}}){ #searches through fragement type keys 
							next if ($match{$ref_name}{$e} =~ /middle/); #skips middle fragments

							#Marks to skip if the hit AAI matches the query AAI > the match AAI does
						
							if ((abs($inblast2{$ref_name}{'AAI'}[$temp_inblast2_best_hit_array_key] - $inblast1{$b}{'AAI'}[$c])) > 
								(abs($inblast2{$ref_name}{'AAI'}[$temp_inblast2_best_hit_array_key] - $match{$ref_name}{$e}{'AAI'}))){
								$next_counter = 1;
							} 

							#Otherwise flags match as better, removes previous indication to skip; this step need because it is possible
							#	to have 2 ends + the query a "middle", and the new end might only be better than one of them
			
							else {
								$temp_best_difference_key{$e} = abs($inblast2{$ref_name}{'AAI'}[$temp_inblast2_best_hit_array_key] - $match{$ref_name}{$e}{'AAI'});
								$next_counter = 0;
							}
						}

						#determines which of the existing matching ends to skip

						unless ($next_counter == 1){
							if (!$temp_best_difference_key{'end1'}){$end_name = "end2"}
							elsif (!$temp_best_difference_key{'end2'}){$end_name = "end1"}
							elsif ($temp_best_difference_key{'end1'} >= $temp_best_difference_key{'end2'}){$end_name = "end1"}
							elsif ($temp_best_difference_key{'end1'} < $temp_best_difference_key{'end2'}){$end_name = "end2"}
						}
					}
		
					#If 2 ends have not been previously found, set the name to add the current end fragment

					else {
						$end_counter++;
						$end_name = "end".$end_counter;
					}
				}

				#Sets the name to add the current middle sequence fragment

				else {
					$middle_counter++;
					$end_name = "middle".$middle_counter;
				}

				#Skips if the end sequence is not being replaced; see above

				next if ($next_counter == 1);

				#Adds fragment data

				$match{$ref_name}{$end_name}{'hitname'} = $inblast1{$b}{'hitname'}[$c];
				$match{$ref_name}{$end_name}{'fragment'} = $inblast2{$d}{'ref_fragment'};
				$match{$ref_name}{$end_name}{'hit_length'} = $inblast1{$b}{'hit_length'}[$c];
				$match{$ref_name}{$end_name}{'hit_alignment'} = $inblast1{$b}{'hit_alignment'}[$c];
				$match{$ref_name}{$end_name}{'hit_identities'} = $inblast1{$b}{'hit_identities'}[$c];
				$match{$ref_name}{$end_name}{'query_start'} = $inblast1{$b}{'query_start'}[$c];
				$match{$ref_name}{$end_name}{'query_end'} = $inblast1{$b}{'query_end'}[$c];
				$match{$ref_name}{$end_name}{'subject_start'} = $inblast1{$b}{'subject_start'}[$c];
				$match{$ref_name}{$end_name}{'subject_end'} = $inblast1{$b}{'subject_end'}[$c];
				$match{$ref_name}{$end_name}{'AAI'} = $inblast1{$b}{'AAI'}[$c];
				$match{$ref_name}{$end_name}{'genome_2_hit'} = $inblast2{$ref_name}{'hitname'}[$temp_inblast2_best_hit_array_key];
			}							

		#Once best hits are recovered, a series of other tests are conducted to ensure fidelity

			#Determines if the query sequence is recovered; otherwise deletes the hash element
			
			$found_ref_name = 0;
			foreach $f (sort keys %{$match{$ref_name}}){
				if ($ref_name eq $match{$ref_name}{$f}{'hitname'}){$found_ref_name = 1}
			}
			if ($found_ref_name == 0){
				delete $match{$ref_name};
				next;
			}
			
			#Determines if the starts and stops for each "assembled" protein overlap by less than 25aa. If they do, reject the assembly.
			#Makes arrays of all starts and stops. With the sorted keys, the query sequence is at the start and its corresponding fragment is at the
			#other end.
 
			@query_starts = @query_ends = ();
			foreach $f (sort keys %{$match{$ref_name}}){ 
				if ($match{$ref_name}{$f}{'query_start'} < $match{$ref_name}{$f}{'query_end'}){ #ifs accommodate for backwards alignment
					push @query_starts, $match{$ref_name}{$f}{'query_start'};
					push @query_ends, $match{$ref_name}{$f}{'query_end'};
				}
				elsif ($match{$ref_name}{$f}{'query_start'} > $match{$ref_name}{$f}{'query_end'}){
					push @query_starts, $match{$ref_name}{$f}{'query_end'};
					push @query_ends, $match{$ref_name}{$f}{'query_start'};
				}
			}
			@temp = sort {$a <=> $b} @query_starts;
			@query_starts = @temp;
			@temp = sort {$a <=> $b} @query_ends;
			@query_ends = @temp;

			#Deletes hash element if only match (i.e. to itself, due to the criteria above) found

			if (!$query_starts[1]){
				delete $match{$ref_name};
				next;
			}

			#Looks for overlaps having less than a maximum of 25 amino acids; if found deletes hash element		

			for ($g = 1; $g < scalar @query_starts; $g++){ 
				if ($query_starts[$g] <= ($query_ends[$g-1] - $aa_overlap)){
					delete $match{$ref_name};
					last;				
				}
			}
		}

				%global_found = ();
		foreach $ref_name (sort keys %match){ #foreach hit
			$no_match = $last_counter = $temp_avg_match_AAI = $temp_avg_ref_AAI = $not_found = $last_counter = $better_counter = 0;
			%found_hit_counters = @temp_ref_AAIs = ();
			$d = ''; 

			#Because sets of matches are compared, good matches can be stored in %global_found out of order. Next if this has happened.

			next if ($global_found{$ref_name});

			#Finds each partner for each match, and determines if that partner also reciprocally found the first match.
			#Counters are used here, because if there are any incongruencies between the matching sets, the series of sets is flagged.

			foreach $f (sort keys %{$match{$ref_name}}){ #foreach part 
				$found = 0;
				$a = $match{$ref_name}{$f}{'hitname'};
				foreach $b (sort keys %{$match{$a}}){ #foreach part of the reciprical
					if ($match{$ref_name}{$f}{'hitname'} eq $match{$a}{$b}{'hitname'}){#check for match
						$found = 1;

						#Prevents matching same hit twice per pairwise genome comparison

						$global_found{$match{$ref_name}{$f}{'hitname'}} = " "; 
					}
				}
				if ($found == 0){#if no match
					$no_match = 1;
				}
			}
			next if ($no_match == 1);

			if ($no_match == 0){#if match

				#evaluates if the current match comprises hits from multiple previously identified sets

				foreach $f (sort keys %{$match{$ref_name}}){#foreach part
					$c = $match{$ref_name}{$f}{'hitname'};
					if ($global_index{$c}){
						$found_hit_counters{$global_index{$c}} = " ";
					}
				}

				#if the current match comprises hits from multiple previously identified sets

				if (scalar keys %found_hit_counters > 1){

					#calculates the average AAI for the match hit set
			
					$temp_avg_match_AAI = 0;
					foreach $f (sort keys %{$match{$ref_name}}){
						$temp_avg_match_AAI += $match{$ref_name}{$f}{'AAI'};
					}
					$temp_avg_match_AAI = $temp_avg_match_AAI / (scalar keys %{$match{$ref_name}});

					#calculates the average AAI of teach previously identified hit set

					foreach $hit_counter (keys %found_hit_counters){
						for $evidence_counter (0..$#{$global_match[$hit_counter]}){
							$temp_avg_ref_AAI = 0;
							for $part_counter (0..$#{$global_match[$hit_counter][$evidence_counter]}){
								$temp_avg_ref_AAI += $global_match[$hit_counter][$evidence_counter][$part_counter]{'AAI'};
							}
							push @temp_ref_AAIs, ($temp_avg_ref_AAI / (scalar @{$global_match[$hit_counter][$evidence]}));
						 	@temp_ref_AAIs = sort {$a <=> $b} @temp_ref_AAIs;
							if ($temp_ref_AAIs[$#temp_ref_AAIs] >= $temp_avg_match_AAI){
								$last_counter = 1;
								last;
							}
						}
						last if ($last_counter == 1);
					}
					last if ($last_counter == 1);

					#if all previously identified hit sets have average AAIs worse than that of the current match set, delete the old sets
					#	referenced in @global_match and %global_index

					foreach $hit_counter (keys %found_hit_counters){
						for $evidence_counter (0..$#{$global_match[$hit_counter]}){
							for $part_counter (0..$#{$global_match[$hit_counter][$evidence_counter]}){
								$global_index{$global_match[$hit_counter][$evidence_counter][$part_counter]{'hitname'}} = ();
							}
						}
						$global_match[$hit_counter] = ();
					}
					next if ($last_counter == 1);
				}

				#if the current match shares parts with a previously found set, ensures that all of the previously found set is included in the
				#	current set

				elsif (scalar keys %found_hit_counters == 1){
					foreach (keys %found_hit_counters){$d = $_}
					for $evidence_counter (0..$#{$global_match[$d]}){
						for $part_counter (0..$#{$global_match[$d][$evidence_counter]}){
							$temp_match = 0;
							foreach $f (keys %{$match{$ref_name}}){
								if ($match{$ref_name}{$f}{'hitname'} eq 
									$global_match[$d][$evidence_counter][$part_counter]{'hitname'}){$temp_match = 1};
							}
							if ($temp_match == 0){
								$not_found = 1;
								$last_counter = 1;
								last;
							}
						}
						last if ($last_counter == 1);
					}

					#if there as a part of the previously found set that is not found in the current match set, the average AAI for both sets are
					#	compared and the best set is carried forward

					if ($not_found == 1){
						$temp_avg_match_AAI = 0;
						foreach $f (keys %{$match{$ref_name}}){
							$temp_avg_match_AAI += $match{$ref_name}{$f}{'AAI'};
						}
						$temp_avg_match_AAI = ($temp_avg_match_AAI / scalar keys %{$match{$ref_name}});
						for $evidence_counter (0..$#{$global_match[$d]}){
							$temp_avg_ref_AAI = 0;
							for $part_counter (0..$#{$global_match[$d][$evidence_counter]}){
								$temp_avg_ref_AAI += $global_match[$d][$evidence_counter][$part_counter]{'AAI'};
							}
							push @temp_ref_AAIs, ($temp_avg_ref_AAI / (scalar @{$global_match[$d][$evidence]}));
						 	@temp_ref_AAIs = sort {$a <=> $b} @temp_ref_AAIs;
							if ($temp_ref_AAIs[$#temp_ref_AAIs] >= $temp_avg_match_AAI){
								$better_counter = 1;
								last;
							}
						}
						last if ($better_counter == 1);

						#if all previously identified hit sets have average AAIs worse than that of the current match set, delete the old sets
						#	referenced in @global_match and %global_index

						for $evidence_counter (0..$#{$global_match[$d]}){
							for $part_counter (0..$#{$global_match[$d][$evidence_counter]}){
								$global_index{$global_match[$d][$evidence_counter][$part_counter]{'hitname'}} = ();
							}
						}
						$global_match[$d] = ();
						$d = '';
					}
				}

				if ($better_counter == 1){
					print "\t\tbetter: At least one avg_ref_AAI > avg_temp_AAI - NEXT!\n";
					next;
				}

				#at this point, if the loop is still running the match set is to be added to @global_match
				#if old @global_match and %global_index values were to be deleted, they already would have been

				#$d is only set if the one $hit_counter is found in %global_index, and only carried through to the step adding the match if new parts
				#	are to be added $global_match[$hit_counter]; otherwise it will remain undefined

				if ($d) {
					$hit_counter = $d;
					$evidence_counter = scalar @{$global_match[$hit_counter]};
				}
				else {
					$hit_counter = scalar @global_match;
					$evidence_counter = 0;
				}

				#adds the data in the %match set to @global_match
	
				$part_counter = -1;
				foreach $f (sort keys %{$match{$ref_name}}){
					$part_counter++;
					%{$global_match[$hit_counter][$evidence_counter][$part_counter]} = (
						'hitname' => $match{$ref_name}{$f}{'hitname'},
						'fragment' => $match{$ref_name}{$f}{'fragment'},
						'AAI' => $match{$ref_name}{$f}{'AAI'},
						'hit_length' => $match{$ref_name}{$f}{'hit_length'},
						'hit_alignment' => $match{$ref_name}{$f}{'hit_alignment'},
						'hit_identities' => $match{$ref_name}{$f}{'hit_identities'},
						'query_start' => $match{$ref_name}{$f}{'query_start'},
						'query_end' => $match{$ref_name}{$f}{'query_end'},
						'subject_start' => $match{$ref_name}{$f}{'subject_start'},
						'subject_end' => $match{$ref_name}{$f}{'subject_end'},
						'genome_2_hit' => $match{$ref_name}{$f}{'genome_2_hit'},
					);

					#if needed, sets the index reference for this part

					$c = $match{$ref_name}{$f}{'hitname'};
					unless ($global_index{$c}){
						$global_index{$c} = $hit_counter;
					} 						
				}
			}
		}	
	}
}

######################################################################################################################
# prints output files
######################################################################################################################

open (OUTFILE1, ">$list_output_file") or die "Cannot open $list_output_file";
open (OUTFILE2, ">$descriptions_output_file") or die "Cannot open $descriptions_output_file";
foreach $hit_counter (0..$#global_match){
	%temp = @temp = ();
	next if (scalar $global_match[$hit_counter] == 0);
	foreach $evidence_counter (0..$#{$global_match[$hit_counter]}){
		foreach $part_counter (0..$#{$global_match[$hit_counter][$evidence_counter]}){
			$temp{$global_match[$hit_counter][$evidence_counter][$part_counter]{'hitname'}} = " ";
		}
	}
	foreach (sort keys %temp){push @temp, $_}
	foreach (@temp){push @{$output_list{$temp[0]}}, $_}
	print OUTFILE2 "Set:\t", join ("\t", @temp);
	print OUTFILE2 "\n";
	foreach $evidence_counter (0..$#{$global_match[$hit_counter]}){
		print OUTFILE2 "\tAlignment $evidence_counter:\n";
		foreach $part_counter (0..$#{$global_match[$hit_counter][$evidence_counter]}){
			print OUTFILE2 "Hit name = $global_match[$hit_counter][$evidence_counter][$part_counter]{'hitname'}\t",
				"AAI = $global_match[$hit_counter][$evidence_counter][$part_counter]{'AAI'}\t",
				"Fragment = $global_match[$hit_counter][$evidence_counter][$part_counter]{'fragment'}\t",
				"Query start = $global_match[$hit_counter][$evidence_counter][$part_counter]{'query_start'}\t",
				"Query end = $global_match[$hit_counter][$evidence_counter][$part_counter]{'query_end'}\t",
				"Subject start = $global_match[$hit_counter][$evidence_counter][$part_counter]{'subject_start'}\t",
				"Subject end = $global_match[$hit_counter][$evidence_counter][$part_counter]{'query_end'}\t",
				"Hit length = $global_match[$hit_counter][$evidence_counter][$part_counter]{'hit_length'}\t",
				"Hit identities = $global_match[$hit_counter][$evidence_counter][$part_counter]{'hit_identities'}\t",
				"Hit alignment = $global_match[$hit_counter][$evidence_counter][$part_counter]{'hit_alignment'}\t",
				"Genome 2 hit = $global_match[$hit_counter][$evidence_counter][$part_counter]{'genome_2_hit'}\n";			
		}
	}
print OUTFILE2 "\n";
}
foreach $temp (sort keys %output_list){
	foreach (@{$output_list{$temp}}){print OUTFILE1 "$_\t"}
	print OUTFILE1 "\n";
}
